#include <kch_complex.h>

#include "spfunc.h"

double gamma(double xx)
{
	return exp(gammln(xx));
}

double frl(double xx) // factorial
{
	return factrl(int(xx));
}

double frln(double xx) // ln factorial
{
	return factln(int(xx));
}

// Chi-Square Probability Function
// P(x2 | v)
double px2v(double x, double v)
{
	return gammp(v/2.0, x/2.0);
}

// Q(x2 | v)
double qx2v(double x, double v)
{
	return gammq(v/2.0, x/2.0);
}

// Exponential Integral, Ei(x)
double expnt(double n, double x)
{
	return expint(int(n), x);
}

// Student's Distribution Probability fucntion
// A(t|v), NRC P. 233 (6.4.9)
double stdp(double t, double v)
{
	return 1.0 - betai(v/2.0, 0.5, v/(v+t*t));
}

// F-Distribution Probability Function
// Q(F | v1, v2) = I(v2/2, v1/2, v2/(v2 + v1*F)), NRC P.234 (6.4.11)
double fdp(double v1, double v2, double f)
{
	return betai(v1*0.5, v2*0.5, v2/(v2 + v1 * f));
}

// Cumulative Binomial Probability Distribution
// P(k, n, p), NRC. P. 234 (6. 4. 12)
double cbp(double k, double n, double p)
{
	return betai(k, n-k+1.0, p);
}

#pragma warning(disable: 4715)

// Bessel function Jn(x)
double bessjn(double n, double x)
{
	if(n == 0.0)
		return bessj0(x);
	else if(n == 1.0)
		return bessj1(x);
	else if (n>= 2.0)
		return bessj(int(n), x);
	else
		NR::nrerror(L"bessjn: invalid argument.");
}

// Bessel function Yn(x), x>0
double bessyn(double n, double x)
{
	if(n==0.0)
		return bessy0(x);
	else if(n == 1.0)
		return bessy1(x);
	else if(n >= 2.0)
		return bessy(int(n), x);
	else
		NR::nrerror(L"bessyn: invalid argument.");
}


// Modified Bessel function In(x), for any real x
double bessin(double n, double x)
{
	if(n==0.0)
		return bessi0(x);
	else if(n == 1.0)
		return bessi1(x);
	else if(n >= 2.0)
		return bessi(int(n), x);
	else
		NR::nrerror(L"bessin: invalid argument.");
}

// Modified Bessel function Kn(x), for positive x
double besskn(double n, double x)
{
	if(n==0.0)
		return bessk0(x);
	else if(n == 1.0)
		return bessk1(x);
	else if(n >= 2.0)
		return bessk(int(n), x);
	else
		NR::nrerror(L"besskn: invalid argument.");
}


#pragma warning(default: 4715)

// Bessel function with factional order
double bessjnu(double nu, double x)
{
	double rj, ry, rjp, ryp;
	bessjy(x, nu, rj, ry, rjp, ryp);

	return rj;
}


// Bessel function with factional order
double bessynu(double nu, double x)
{
	double rj, ry, rjp, ryp;
	bessjy(x, nu, rj, ry, rjp, ryp);

	return ry;
}

// Bessel function with factional order
double bessjnup(double nu, double x)
{
	double rj, ry, rjp, ryp;
	bessjy(x, nu, rj, ry, rjp, ryp);

	return rjp;
}

// Bessel function with factional order
double bessynup(double nu, double x)
{
	double rj, ry, rjp, ryp;
	bessjy(x, nu, rj, ry, rjp, ryp);

	return ryp;
}

// Modified Bessel function, Iv(x)
double bessinu(double nu, double x)
{
	double ri, rk, rip, rkp;
	bessik(x, nu, ri, rk, rip, rkp);

	return ri;
}

// Modified Bessel function, Kv(x)
double bessknu(double nu, double x)
{
	double ri, rk, rip, rkp;
	bessik(x, nu, ri, rk, rip, rkp);

	return rk;
}

// Modified Bessel function, derivative of Iv(x)
double bessinup(double nu, double x)
{
	double ri, rk, rip, rkp;
	bessik(x, nu, ri, rk, rip, rkp);

	return rip;
}

// Modified Bessel function, derivative of Kv(x)
double bessknup(double nu, double x)
{
	double ri, rk, rip, rkp;
	bessik(x, nu, ri, rk, rip, rkp);

	return rkp;
}

// Gamma by Chebyshev expansion, P 250
double gamma1 (double x)
{
	double gam1, gam2, gampl, gammi;

	beschb(x, gam1, gam2, gampl, gammi);

	return gam1;
}

// Gamma by Chebyshev expansion, P 250
double gamma2 (double x)
{
	double gam1, gam2, gampl, gammi;

	beschb(x, gam1, gam2, gampl, gammi);

	return gam2;
}

// Gamma by Chebyshev expansion, P 250
double gampl (double x)
{
	double gam1, gam2, gampl, gammi;

	beschb(x, gam1, gam2, gampl, gammi);

	return gampl;
}

// Gamma by Chebyshev expansion, P 250, 1/gamma(1+x)
double gammi (double x)
{
	double gam1, gam2, gampl, gammi;

	beschb(x, gam1, gam2, gampl, gammi);

	return gammi;
}

// Airy function, Ai(x)
double ai(double x)
{
	double ai, bi, aip, bip;

	airy(x, ai, bi, aip, bip);

	return ai;
}

// Airy function, Bi(x)
double bi(double x)
{
	double ai, bi, aip, bip;

	airy(x, ai, bi, aip, bip);

	return bi;
}

// Airy function, derivative of Ai(x)
double aip(double x)
{
	double ai, bi, aip, bip;

	airy(x, ai, bi, aip, bip);

	return aip;
}

// Airy function, derivative of Bi(x)
double bip(double x)
{
	double ai, bi, aip, bip;

	airy(x, ai, bi, aip, bip);

	return bip;
}

// Spherical Bessel Functions, jn(x)
double spjn(double n, double x)
{
	double sj, sy, sjp, syp;

	sphbes(int(n), x, sj, sy, sjp, syp);

	return sj;
}

// Spherical Bessel Functions, yn(x)
double spyn(double n, double x)
{
	double sj, sy, sjp, syp;

	sphbes(int(n), x, sj, sy, sjp, syp);

	return sy;
}


// Spherical Bessel Functions, derivative of jn(x)
double spjnp(double n, double x)
{
	double sj, sy, sjp, syp;

	sphbes(int(n), x, sj, sy, sjp, syp);

	return sjp;
}

// Spherical Bessel Functions, derivative of yn(x)
double spynp(double n, double x)
{
	double sj, sy, sjp, syp;

	sphbes(int(n), x, sj, sy, sjp, syp);

	return syp;
}

// Legendre Polynomial, P. 258
double legp(double l, double m, double x)
{
	return plgndr(int(l), int(m), x);
}

// Frenel integral C(x)
double frec(double x)
{
	Complex cs;
	frenel(x, cs);

	return cs.real();
}


// Frenel integral S(x)
double fres(double x)
{
	Complex cs;
	frenel(x, cs);

	return cs.imag();
}


// Cosine integeral, Ci(x)
double ci(double x)
{
	Complex cs;
	cisi(x, cs);

	return cs.real();
}


// Sine integeral, Si(x)
double si(double x)
{
	Complex cs;
	cisi(x, cs);

	return cs.imag();
}

// Jacobian elliptic fucntions sn(u, kc)
double sn(double u, double kc)
{
	double emmc = kc * kc, sn, cn, dn;
	sncndn(u, emmc, sn, cn, dn);

	return sn;
}


// Jacobian elliptic fucntions cn(u, kc)
double cn(double u, double kc)
{
	double emmc = kc * kc, sn, cn, dn;
	sncndn(u, emmc, sn, cn, dn);

	return cn;
}

// Jacobian elliptic fucntions dn(u, kc)
double dn(double u, double kc)
{
	double emmc = kc * kc, sn, cn, dn;
	sncndn(u, emmc, sn, cn, dn);

	return dn;
}

// Hypergeometirc series 2F1, P. 276
Complex hypsr (const Complex& a, const Complex& b,
			   const Complex& c, const Complex& z)
{
	Complex series, deriv;

	hypser(a, b, c, z, series, deriv);

	return series;
}

// Hypergeometirc series 2F1's derivative, P. 276
Complex hypsrd (const Complex& a, const Complex& b,
			   const Complex& c, const Complex& z)
{
	Complex series, deriv;

	hypser(a, b, c, z, series, deriv);

	return deriv;
}

