#pragma once

#include <vector>
#include <algorithm>
#include <hash_map>
#include <deque>
#include <string>

#include <kch_basetype.h>

#include <kch_exception.h>
#include <kch_vector.h>
#include <kch_complex.h>

#include <afxwin.h>
KCH_EXP void PutString(const char*);

template<typename T>
class ptrvector: public std::vector<T*>
{
protected:
	bool deleted;
	void clearall()
	{
		if(deleted==false)
		{
			
			for(std::vector<T*>::size_type i = 0; i < this->size(); ++i)
			{
				T* ptr= this->at(i);
				if(ptr != NULL)
				{
					delete ptr; ptr = NULL;
				}
			}
			

			this->clear();
			deleted = true;
		}
	}
public:
	ptrvector(T* tptr=NULL): deleted(false)
	{
		if(tptr != NULL) this->push_back(tptr);
	}

	void push_back(T* pVal)
	{
		if(pVal)
		{
			std::vector<T*>::push_back(pVal); deleted=false;
		}
	}

	void Delete() { clearall(); }
	~ptrvector() { clearall(); }
};

template<typename T>
class Stack
{
protected:
	// container for the elements
	std::deque<T> c; 
public:
	/* exception class for pop() and top() with empty stack */

	class stack_empty: public std::exception 
	{
	public:
		virtual const char* what() const throw()
		{
			return "read empty stack";
		}
	};

	// number of elements
	typename std::deque<T>::size_type size() const
	{
		return c.size();
	}

	// is stack empty?
	bool empty() const 
	{
		return c.empty();
	}

	// push element into the stack
	void push(const T& elem)
	{
		c.push_back(elem);
	}

	// pop element out of the stack and return its value
	T pop()
	{
		if(c.empty())
		{
			throw stack_empty();
		}

		T elem(c.back());
		c.pop_back();

		return elem;
	}

	T bottom()
	{
		if(c.empty())
			throw stack_empty();
		
		T elem(c[0]);
		return elem;
	}

	void clear() { c.clear(); }
	void Cleanup()
	{
		T elem;
		while(!c.empty())
		{
			elem = c.back();
			delete elem;
			c.pop_back();
		}
	}

	T getoffset(int offset)
	{
		int index = int(c.size())-offset;

		if(index<0) 
			return NULL;
		else
			return c[index];
	}

	// return the value of next element
	T& top()
	{
		if(c.empty())
		{
			throw stack_empty();
		}
		return c.back();
	}
};

enum DataType { TypeNull, TypeReal, TypeCplx, TypeVctr};

class Data: public BaseType
{
protected:
	DataType dtype;
	string id;

public:
	Data(string varid, DataType t = TypeNull):
	  id(varid), dtype(t) {/* empty */ }
	virtual DataType GetType() const { return dtype; }
	virtual const string& GetId() const { return id; }

	virtual ~Data() { /* empty */ }
};

class RealData: public Data
{

protected:
	double data;

public:
	RealData(string varid, double r=0): 
	  data(r), Data(varid, TypeReal)
	{ /* empty */ }

	double& GetData() { return data; }

	virtual ~RealData() { /* empty */ }
};

class CplxData: public Data
{

protected:
	Complex data;

public:
	CplxData(string varid="", double r=0, double i=0): 
	  data(MakeComplex(r, i)), Data(varid, TypeCplx)
	{ /* empty */ }

	Complex& GetData() { return data; }

	virtual ~CplxData() { /* empty */ }
};

class VctrData: public Data
{
protected:
	Vector data;

public:
	VctrData(string varid = "", 
		double i=0, double j=0, double k=0, bool is_scalar=false): 
	  data(MakeVector(i, j, k, is_scalar)), Data(varid, TypeVctr)
	{	/* empty */  }

	Vector& GetData() { return data; }

	virtual ~VctrData() { /* empty */ }
};

class VarList
{
public:
	typedef pair<const string, int> IndexPair;
	typedef stdext::hash_map<const string, int>::iterator IndexItr;

protected:
	stdext::hash_map<const string, int> varindex;
	std::vector<Data*> varlist;
public:

	int insert(Data* pData)
	{
		IndexItr itr = varindex.find(pData->GetId());
		
		if(itr != varindex.end()) 
			return itr->second;
		else
		{
			varlist.push_back(pData);
			int index = int(varlist.size()) - 1;
			varindex.insert(IndexPair(pData->GetId(), index));
			return index;
		}
	}

	Data* operator[](const string& id)
	{
		IndexItr itr = varindex.find(id);
		if(itr != varindex.end()) 
		{
			return varlist[itr->second];
		}
		else
		{			
			throw bad_index("Variable::operator[](const string& id): id out of range");
		}
	}

	DataType GetType(const string& id)
	{
		IndexItr itr = varindex.find(id);
		if(itr != varindex.end()) 
		{
			return varlist[itr->second]->GetType();
		}
		else
			return TypeNull;
	}

	Data* operator[](int index)
	{
		if(index < int(varlist.size()))
			return varlist[index];
		else
			throw bad_index("Variable::operator[](int index): index out of range");
	}

	void Delete()
	{
		//varlist.Delete();
		varindex.clear();
	}

	void PrintElement()
	{
		std::ostringstream os;

		for(int i=0; i<int(varlist.size()); ++i)
		{
			os<<varlist[i]->GetId()<<" ";
		}
		os<<std::endl;
		
		PutString(os.str().c_str());
	}
};
