#pragma once

#ifdef LIBBUILD
#define KCH_EXP __declspec(dllexport)
#else
#define KCH_EXP __declspec(dllimport)
#endif

#include <kch_exception.h>
#include <kch_complex.h>
#include <kch_vector.h>
#include <kch_basetype.h>
#include <kch_etl.h>

#include <cstring>
#include <string>
#include <sstream>

#ifndef VARNAME_LENGTH
#define VARNAME_LENGTH	128
#endif

KCH_EXP void yyerror(char const*);
KCH_EXP int yylex();
KCH_EXP bool yyinitialize();
KCH_EXP void yycleanup();
KCH_EXP void Cleanup();
KCH_EXP void PutString(const char*);

extern char *scanbuffer;
extern int scanbuffer_size;
extern int scanbuffer_fsize;
extern int lineno, charcount;

template<typename T>
class Expression: public BaseType
{
protected:
	Expression() { /* not to be instantiated */ }

public:
	virtual T Eval() { throw not_supported("Eval() not supported"); }
	virtual bool EvalBool() { throw not_supported("EvalBool() not supported"); }
	virtual std::string Display() = 0;

	virtual ~Expression() { /* empty */}
};

template<typename T, typename U> 
class Variable: public Expression<T>
{
public:
	string varid;
	Variable(string id): varid(id) { }

	virtual T Eval() 
	{
		Data* pData=LookupData(varid);

		if(pData != NULL)
		{
			return dynamic_cast<U*>(pData)->GetData();
		}
		else
			// this can not happen;
			throw exception("Variable<double>::Eval(), Variable not found.");;
	}

	virtual bool EvalBool() 
	{ throw not_supported("Variable ::EvalBool(): method not supported"); }
	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<varid;
		return ostr.str();
	}
	
	virtual ~Variable() { }
};

template<typename T> 
class Literal: public Expression<T>
{
protected:
	T val;
public:
	Literal(T v=T()): val(v) { }

	virtual T Eval() { return val; }
	virtual bool EvalBool() 
	{ throw not_supported("Literal::EvalBool(): method not supported"); }
	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<val<<endl;
		return ostr.str();
	}
	
	virtual ~Literal() { }
};

template<> 
class Literal<Complex>: public Expression<Complex>
{
protected:
	Expression<double>* lhs, *rhs;
public:
	Literal(Expression<double>* l, Expression<double>* r=NULL): 
	  lhs(l), rhs(r) { }

	virtual Complex Eval() { return MakeComplex(lhs->Eval(), (rhs == NULL) ? 0: rhs->Eval());}
	virtual bool EvalBool() 
	{ throw not_supported("Literal::EvalBool(): method not supported"); }
	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<MakeComplex(lhs->Eval(), (rhs == NULL) ? 0: rhs->Eval());
		return ostr.str();
	}
	
	virtual ~Literal() { }
};

template<> 
class Literal<Vector>: public Expression<Vector>
{
protected:
	Expression<double>* x, *y, *z;
public:
	Literal(Expression<double>* i, Expression<double>* j, Expression<double>* k): 
	  x(i), y(j), z(k) { }

	virtual Vector Eval() { return MakeVector(x->Eval(), y->Eval(), z->Eval() );}
	virtual bool EvalBool() 
	{ throw not_supported("Literal::EvalBool(): method not supported"); }
	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<MakeVector(x->Eval(), y->Eval(), z->Eval() );
		return ostr.str();
	}
	
	virtual ~Literal() { }
};

template<typename T> 
class UnaryOpr: public Expression<T>
{
protected:
	Expression<T> *expr;
	int optr; // operator

public:
	UnaryOpr(int opr, Expression<T>* xpr): optr(opr), expr(xpr) { }

	virtual T Eval()
	{
		switch(optr)
		{
		case '-': return -expr->Eval();
		default:
			throw not_supported("UnaryOpr::Eval(): operator not supported");
		}
	}

	virtual bool EvalBool() 
	{
		switch(optr)
		{
		case '!': return !expr->EvalBool();

		default:
			throw not_supported("UnaryOpr::EvalBool(): operator not supported");
		}
	}

	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<'('<<char(optr)<<expr->Display()<<')';
		return ostr.str();
	}

	virtual ~UnaryOpr() { }
};

template<typename T> 
class BinOpr: public Expression<T>
{
protected:
	int optr; // operator
	Expression<T> *lhs, *rhs;
	
public:
	BinOpr(int opr, Expression<T>* ls, Expression<T>* rs):
	  optr(opr), lhs(ls), rhs(rs) { /* empty */ }
	
	virtual T Eval() 
	{ 
		switch(optr)
		{
		case '+': return lhs->Eval() + rhs->Eval();
		case '-': return lhs->Eval() - rhs->Eval();
		case '*': return lhs->Eval() * rhs->Eval();
		case '/': return lhs->Eval() / rhs->Eval();
		default:
			throw not_supported("BinOpr::Eval(): operator not supported");
		}
	}
	
	virtual bool EvalBool() 
	{ 
		switch(optr)
		{
		case '>': return lhs->Eval() > rhs->Eval();
		case '<': return lhs->Eval() < rhs->Eval();
		case '(': return lhs->Eval() <= rhs->Eval();
		case ')': return lhs->Eval() >= rhs->Eval();
		case '$': return lhs->Eval() == rhs->Eval();
		case '!': return lhs->Eval() != rhs->Eval();
		 
		default:
			throw not_supported("BinOpr::Eval(): operator not supported");
		}
	}

	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<'('<<lhs->Display()<<char(optr)<<rhs->Display()<<')';
		return ostr.str();
	}

	virtual ~BinOpr() { }
};

class ModOpr: public BinOpr<double>
{
public:
	ModOpr(Expression<double>* ls, Expression<double>* rs):
	  BinOpr('%', ls, rs) { /* empty */ }
	
	virtual double Eval() 
	{ 
		return double(long(lhs->Eval()) % long(rhs->Eval()));
	}
	
	virtual ~ModOpr() { }
};

template<> 
class BinOpr<Vector>: public Expression<Vector>
{
protected:
	int optr; // operator
	Expression<Vector> *lhs, *rhs;
	
public:
	BinOpr(int opr, Expression<Vector>* ls, Expression<Vector>* rs):
	  optr(opr), lhs(ls), rhs(rs) { }
	
	virtual Vector Eval() 
	{ 
		switch(optr)
		{
		case '+': return lhs->Eval() + rhs->Eval();
		case '-': return lhs->Eval() - rhs->Eval();
		case '%': return lhs->Eval() % rhs->Eval(); // cross product
		
		default:
			throw not_supported("BinOpr::Eval(): operator not supported");
		}
	}
	
	virtual bool EvalBool() 
	{ 
		switch(optr)
		{
		case '>': return lhs->Eval() > rhs->Eval();
		case '<': return lhs->Eval() < rhs->Eval();
		case '(': return lhs->Eval() <= rhs->Eval();
		case ')': return lhs->Eval() >= rhs->Eval();
		case '$': return lhs->Eval() == rhs->Eval();
		case '!': return lhs->Eval() != rhs->Eval();
		 
		default:
			throw not_supported("BinOpr::Eval(): operator not supported");
		}
	}

	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<'('<<lhs->Display()<<char(optr)<<rhs->Display()<<')';
		return ostr.str();
	}

	virtual ~BinOpr() { }
};

// return vector
// scalar op vector
class RVBinOpr: public BinOpr<Vector>
{
protected:
	Expression<double>* s;
	
public:
	RVBinOpr(int opr, Expression<double>* scalar, Expression<Vector>* vctr):
	  s(scalar), BinOpr(opr, NULL, vctr) { }
   
	RVBinOpr(int opr, Expression<Vector>* vctr, Expression<double>* scalar):
	  s(scalar), BinOpr(opr, NULL, vctr) { }
	 
    virtual Vector Eval()
	{
		switch(optr)
		{
		case '*':
			return s->Eval() * rhs->Eval();
		case '/':
			return rhs->Eval() / s->Eval();
		
		default:
			throw exception("Operation not supported\r\n");
		}
	}
	
	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<'('<<s->Display()<<char(optr)<<rhs->Display()<<')';
		return ostr.str();
	}

	virtual ~RVBinOpr() { }
};

// vector dot product
class VVDotOpr: public BinOpr<double>
{
protected:
	Expression<Vector>* lhs;
	Expression<Vector>* rhs;

public:
	VVDotOpr(Expression<Vector>* left, Expression<Vector>* right):
	  lhs(left), rhs(right), BinOpr('*', NULL, NULL) { }
    virtual double Eval()
	{
		return lhs->Eval() * rhs->Eval();
	}
	
	virtual std::string Display() 
	{
		std::ostringstream ostr;
		ostr<<'('<<lhs->Display()<<'*'<<rhs->Display()<<')';
		return ostr.str();
	}
	
	virtual bool EvalBool() 
	{ 
		throw not_supported("BinOpr::Eval(): operator not supported");
	}

	virtual ~VVDotOpr() { }
};

class Statement: public BaseType
{
protected:
	std::string m_msg;
public:
	Statement(std::string txt=""): m_msg(txt) {  }
	virtual int Run() { PutString(m_msg.c_str()); return 0; }
	virtual bool IsLabel(std::string label="") { return false; }
	virtual bool IsLoop() { return false; }
	virtual ~Statement() { }
};

class Label: public Statement
{
public:
	Label(std::string txt): Statement(txt) {  }
	virtual int Run() 
	{ 
		return 0; 
	}
	virtual bool IsLabel(std::string label="") 
	{ 
		return m_msg == label;
	}
	virtual ~Label() { }
};

template<typename T, typename U>
class Declare: public Statement
{
public:
	Variable<T, U>* pVar;
	Expression<T>* pExpr;

	Declare(Variable<T, U>* var, Expression<T>* expr = NULL): 
	pVar(var), pExpr(expr) { }
	virtual int Run()
	{
		U* pValue = NULL;			
		if(pExpr != NULL)
		{
			if(pValue = dynamic_cast<U*>(LookupData(pVar->varid))) 
				pValue->GetData() = pExpr->Eval();
		}
		return 0;
	}
	virtual ~Declare() { }
};

template<typename T, typename U>
class Assignment: public Statement
{
public:
	Variable<T, U>* pVar;
	Expression<T>* pExpr;
	Assignment(Variable<T, U>* var, Expression<T>* xpr):
	  pVar(var), pExpr(xpr) { }
	
	virtual int Run()
	{
		//extern VarList* gVarList;
		Data* pData = LookupData(pVar->varid);
		if(pData == NULL) return 0;

		U* pValue = dynamic_cast<U*>(pData);
		pValue->GetData() = pExpr->Eval();
		
		return 0;
	}
	
	virtual ~Assignment() { }
};

template<typename T>
class StmtIf: public Statement
{
protected:
	Expression<T>* pExpr;
	Statement *pStmt;

public:
	StmtIf(Expression<T>* expr, Statement* stmt): 
	  pExpr(expr), pStmt(stmt) { }
	virtual int Run()
	{
		if(pExpr->EvalBool())
			return pStmt->Run();
		return 0;
	}
	
	virtual ~StmtIf() { }
};

template<typename T>
class StmtIfElse: public Statement
{
protected:
	Expression<T>* pExpr;
	Statement *pStmt, *pStmtElse;

public:
	StmtIfElse(Expression<T>* expr, Statement* stmt, Statement* stmtelse): 
	  pExpr(expr), pStmt(stmt), pStmtElse(stmtelse) { }
	virtual int Run()
	{
		if(pExpr->EvalBool())
			return pStmt->Run();
		else
			return pStmtElse->Run();
	}

	virtual ~StmtIfElse() { }
};

template<typename T>
class StmtWhile: public Statement
{
protected:
	Expression<T>* pExpr;
	Statement *pStmt;

public:
	StmtWhile(Expression<T>* expr, Statement* stmt): 
	  pExpr(expr), pStmt(stmt) { }
	virtual int Run()
	{
		extern Stack<StatementList*> gStmtListStack;
		StatementList* pStackTop = gStmtListStack.top();

		while(pExpr->EvalBool())
		{
			try
			{
				pStmt->Run();
			}
			catch(Break::BreakLoop&)
			{
				while(pStackTop != gStmtListStack.top())
				{
					gStmtListStack.pop();
					//PutString("}\r\n");
				};
				break;
			}
			catch(Skip::SkipLoop&)
			{
				while(pStackTop != gStmtListStack.top())
				{
					gStmtListStack.pop();
					//PutString("}\r\n");
				};
			}
		}
		return 0;
	}

	virtual bool IsLoop() { return true; }
	virtual ~StmtWhile() { }
};

template<typename T>
class StmtFor: public Statement
{
protected:
	Expression<T>* expr;
	Statement* stmt2;
	Statement *stmt;

public:
	Statement* stmt1;

	StmtFor(Statement* st1, Expression<T>* xpr, Statement* st2, Statement* st):
		stmt1(st1), expr(xpr), stmt2(st2), stmt(st) { }

	virtual int Run()
	{
		extern Stack<StatementList*> gStmtListStack;
		StatementList* pStackTop = gStmtListStack.top();

		for(stmt1->Run(); expr->EvalBool(); stmt2->Run())
		{
			try
			{
				stmt->Run();
			}
			catch(Break::BreakLoop&)
			{
				while(pStackTop != gStmtListStack.top())
				{
					gStmtListStack.pop();
					//PutString("}\r\n");
				};
				break;
			}
			catch(Skip::SkipLoop&)
			{
				while(pStackTop != gStmtListStack.top())
				{
					gStmtListStack.pop();
					//PutString("}\r\n");
				};
			}
		}

		return 0;
	}
	
	virtual bool IsLoop() { return true; }
	virtual ~StmtFor() { }
};

class GoTo: public Statement
{
public:
	GoTo(std::string txt): Statement(txt) { }

	class GoToLabel: public std::exception
	{
	public:

		GoToLabel(const char* txt="Toto Label"): exception(txt) {}
	};

	virtual int Run()
	{
		throw GoToLabel(m_msg.c_str());
	}
	
	virtual ~GoTo() { }
};

class Skip: public Statement
{
public:
	Skip() { }

	class SkipLoop: public std::exception
	{
	public:
		SkipLoop(const char* msg="Break Loop"): exception(msg) {}
	};

	virtual int Run()
	{
		throw SkipLoop();
	}
	
	virtual ~Skip() { }
};

template<typename T>
class ExprType: public Statement
{
protected:
	Expression<T>* expr;

public:
	ExprType(Expression<T>* xpr): expr(xpr) { }
	virtual int Run()
	{ 
		std::ostringstream ostr;
		ostr<<expr->Display()<<"="<<expr->Eval()<<"\r\n"; 
		PutString(ostr.str().c_str());

		return 0;
	}
	virtual ~ExprType() { }
};

extern ptrvector<BaseType> gAllocMem;

class StatementList: public Statement
{
protected:
	std::vector<Statement*> stmtlist;
	VarList varlist;

public:
	StatementList(){  }

	VarList* GetVarList() { return &varlist; }

	void Add(Statement* stmt)
	{
		if(stmt != NULL)
		{
			stmtlist.push_back(stmt);
						
			Data *pData=NULL;

			Declare<double, RealData> *pReal = 
				dynamic_cast<Declare<double, RealData>*>(stmt);
			if(pReal != NULL)
				pData = new RealData(pReal->pVar->varid);
						
			Declare<Complex, CplxData> *pCplx = 
				dynamic_cast<Declare<Complex, CplxData>*>(stmt);
			if(pCplx != NULL)
				pData = new CplxData(pCplx->pVar->varid);
						
			Declare<Vector, VctrData> *pVctr = 
				dynamic_cast<Declare<Vector, VctrData>*>(stmt);
			if(pVctr != NULL)
				pData = new VctrData(pVctr->pVar->varid);
						
			if(pData != NULL) 
			{
				gAllocMem.push_back(pData);
				varlist.insert(pData);
			}
		}
	}
	
	virtual int Run()
	{
		std::vector<Statement*>::size_type size = stmtlist.size();
		
		extern Stack<StatementList*> gStmtListStack;
		
		//PutString("{\r\n");
		for(std::vector<Statement*>::size_type i=0; i<size; ++i)
		{
			gStmtListStack.push(this);
			
			try
			{
				stmtlist[i]->Run();
			}
			catch(GoTo::GoToLabel& id)
			{
				int index = FindLabel(std::string(id.what()));
				if(index >= 0) 
					i = std::vector<Statement*>::size_type(index);
				else
				{
					gStmtListStack.pop();
					//PutString("}\r\n");
					throw;
				}
			}
			
			gStmtListStack.pop();
		}
		//PutString("}\r\n");

		return 0;
	}

	// if fails, returns -1, otherwise returns stmtlist index
	int FindLabel(std::string& id)
	{
		std::vector<Statement*>::size_type size = stmtlist.size();
		for(std::vector<Statement*>::size_type i=0; i<size; ++i)
		{
			if(stmtlist[i]->IsLabel(id)) return int(i);
		}
	
		return -1;
	}

	// if fails, returns -1, otherwise returns stmtlist index
	bool FindLoop()
	{
		std::vector<Statement*>::size_type size = stmtlist.size();
		for(std::vector<Statement*>::size_type i=0; i<size; ++i)
		{
			if(stmtlist[i]->IsLoop()) return true;
		}
	
		return false;
	}

	virtual ~StatementList() { }
};


inline DataType LookupDataType(const std::string id)
{
	extern Stack<StatementList*> gStmtListStack;

	StatementList* pStatementList = gStmtListStack.top();
	VarList* pVarList = pStatementList->GetVarList();
		
	DataType type= pVarList->GetType(id);
	if(type != TypeNull) return type;

	for(int i=1; ;++i)
	{
		pStatementList = gStmtListStack.getoffset(i);
		
		if(pStatementList)
		{
			// pVarList is always not NULL
			pVarList = pStatementList->GetVarList();
	
			type = pVarList->GetType(id);
			if(type != TypeNull) return type;
		}
		else return TypeNull;
	}
}

inline Data* LookupData(const std::string id)
{
	extern Stack<StatementList*> gStmtListStack;

	StatementList* pStatementList = gStmtListStack.top();
	VarList* pVarList = pStatementList->GetVarList();
		
	DataType type= pVarList->GetType(id);
	if(type != TypeNull)
		return pVarList->operator[](id);
	
	for(int i=1;; ++i)
	{
		pStatementList = gStmtListStack.getoffset(i);

		if(pStatementList)
		{
			// pVarList is always not NULL
			pVarList = pStatementList->GetVarList();
			
			type = pVarList->GetType(id);
			if(type != TypeNull)
				return pVarList->operator[](id);
		}
		else return NULL;
	}
}

inline bool LookupLoop()
{
	extern Stack<StatementList*> gStmtListStack;

	StatementList* pStatementList = NULL;
	
	for(int i=1; ;++i)
	{
		pStatementList = gStmtListStack.getoffset(i);
		if(pStatementList)
		{
			if(pStatementList->FindLoop()) 
				return true;
			else
				continue;
		}
		else return false;
	}
}

class Break: public Statement
{
public:
	Break() { }
	class BreakLoop: public std::exception
	{
	public:

		BreakLoop(const char* msg="Break Loop"): exception(msg) {}
	};

	virtual int Run()
	{
		if(LookupLoop()) 
			throw BreakLoop();
		return 0;
	}
	
	virtual ~Break() { }
};