#pragma once

#include <cmath>
#include <iostream>
#include <ios>

template <typename T>
class tmp_vec
{
protected:
	

public:
	bool scalar; T x, y, z;

	tmp_vec(T i=T(), T j=T(), T k=T(), bool s=false):
	  x(i), y(j), z(k), scalar(s) { }

	tmp_vec(const tmp_vec& v)
	{
		x = v.x; y = v.y; z = v.z; scalar = v.scalar;
	}

	tmp_vec& operator = (const tmp_vec& v)
	{
		if(this != &v)
		{
			x = v.x; y = v.y; z = v.z; scalar = v.scalar;
		}

		return *this;
	}

	tmp_vec operator+(const tmp_vec& v) const
	{
		return tmp_vec(x+v.x, y+v.y, z+v.z);
	}

	tmp_vec operator-(const tmp_vec& v) const
	{
		return tmp_vec(x-v.x, y-v.y, z-v.z);
	}

	tmp_vec& operator+=(const tmp_vec& v)
	{
		x += v.x; y += v.y; z += v.z;
		return *this;
	}

	tmp_vec& operator-=(const tmp_vec& v)
	{
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}

	tmp_vec& operator*=(const tmp_vec& v)
	{
		throw UnicodeException(_T("Operation not defined"));
	}

	tmp_vec& operator/=(const tmp_vec& v)
	{
		throw UnicodeException(_T("Operation not defined"));
	}

	tmp_vec& operator+=(T v)
	{
		x += v; y += v; z += v;
		return *this;
	}

	tmp_vec& operator-=(T v)
	{
		x -= v; y -= v; z -= v;
		return *this;
	}

	tmp_vec& operator*=(T v)
	{
		x *= v; y *= v; z *= v;
		return *this;
	}

	tmp_vec& operator/=(T v)
	{
		if(v == T()) 
			throw UnicodeException(_T("Division by Zero in vector operation"));

		x /= v; y /= v; z /= v;
		return *this;
	}

	// unary minus
	tmp_vec operator-() 
	{
		return tmp_vec(-x, -y, -z);
	}

	// cross product
	tmp_vec operator %(const tmp_vec& v) const
	{
		return tmp_vec(y*v.z-z*v.y, z*v.x-x*v.z, x*v.y-y*v.x);
	}

	tmp_vec& operator%=(const tmp_vec& v)
	{
		tmp_vec vec(y*v.z-z*v.y, z*v.x-x*v.z, x*v.y-y*v.x);
		*this = vec; return *this;
	}

	// dot product
	T operator*(const tmp_vec& v) const
	{
		return x*v.x+y*v.y+z*v.z;
	}

	// scalar * vector
	friend tmp_vec operator*(const tmp_vec& v, T d) 
	{
		return tmp_vec(d*v.x, d*v.y, d*v.z);
	}

	// vector * scalar
	friend tmp_vec operator*(T d, const tmp_vec& v)
	{
		return tmp_vec(v.x*d, v.y*d, v.z*d);
	}

	// vector / scalar
	friend tmp_vec operator/(const tmp_vec& v, T d)
	{
		if(d == T()) throw UnicodeException(_T("Division by Zero"));

		return tmp_vec(v.x/d, v.y/d, v.z/d);
	}

	// absolute value
	friend T absv(const tmp_vec& v)
	{
		return sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
	}

	// norm value
	friend T normv(const tmp_vec& v)
	{
		return v.x*v.x+v.y*v.y+v.z*v.z;
	}
	
	friend bool operator > (const tmp_vec& v1, const tmp_vec& v2)
	{
		return normv(v1)>normv(v2);
	}

	friend bool operator < (const tmp_vec& v1, const tmp_vec& v2)
	{
		return normv(v1)<normv(v2);
	}

	friend bool operator == (const tmp_vec& v1, const tmp_vec& v2)
	{
		return (v1.x==v2.x)&&(v1.y==v2.y)&&(v1.z==v2.z);
	}

	friend bool operator != (const tmp_vec& v1, const tmp_vec& v2)
	{
		return !(v1==v2);
	}

	friend bool operator >= (const tmp_vec& v1, const tmp_vec& v2)
	{
		double n1 = normv(v1), n2 = normv(v2);
		if(n1>n2)
			return true;
		else if(v1 == v2)
			return true;
		else
			return false;
	}

	friend bool operator <= (const tmp_vec& v1, const tmp_vec& v2)
	{
		T n1 = normv(v1), n2 = normv(v2);
		if(n1<n2)
			return true;
		else if(v1 == v2)
			return true;
		else
			return false;
	}

	friend T xcomv(const tmp_vec& v)
	{
		return v.x;
	}

	friend T ycomv(const tmp_vec& v)
	{
		return v.y;
	}

	friend T zcomv(const tmp_vec& v)
	{
		return v.z;
	}

	friend tmp_vec getunitvector( const tmp_vec& v)
	{
		T  mag = absv(v);
		
		return v/mag;
	}

	// cylindrical to rectangular
	friend tmp_vec c2rv(T rho, T psi, T z)
	{
		return tmp_vec(rho*cos(psi), rho*sin(psi), z); 
	}

	// rectangler to cylindrical 
	friend tmp_vec r2cv(T x, T y, T z)
	{
		T p = sqrt(x*x+y*y);
		T psi;
		T mpi2 = M_PI/2.0;
		
		if(p == T())
			psi = T();
		else
		{
			if(x==T())
			{
				if(y>T())
					psi = mpi2;
				else // y<0.0
					psi = -mpi2;
			}
			else
				psi = atan(y/x);
		}

		return tmp_vec(p, psi, z);
	}

	// spherical to rectanglur
	friend tmp_vec s2rv(T r, T theta, T psi)
	{
		T p = r*sin(theta);
		return tmp_vec(p*cos(psi), p*sin(psi), r*cos(theta));
	}

	// rectangular to spherical
	friend tmp_vec r2sv(T x, T y, T z)
	{
		T p = x*x+y*y;
		T r = sqrt(p+z*z);
		T theta;
		T psi;

		T mpi2 = M_PI/2.0;

		if(z == T())
			theta = mpi2;
		else
			theta = atan(sqrt(p)/z);

		if(x == T())
		{
			if(y>T())
				psi = mpi2;
			else if(y<T())
				psi = -mpi2;
			else
				psi = T();
		}
		else
			psi = atan(y/x);

		return tmp_vec(r, theta, psi);
	}


	// cylindrical to rectangular
	friend tmp_vec c2rvv(const tmp_vec& v)
	{
		return c2rv(v.x, v.y, v.z);
	}

	// cylindrical to rectangler
	friend tmp_vec r2cvv(const tmp_vec& v)
	{
		return r2cv(v.x, v.y, v.z);
	}

	// sperical to rectanglur
	friend tmp_vec s2rvv(const tmp_vec& v)
	{
		return s2rv(v.x, v.y, v.z);
	}

	// rectangular to spherical
	friend tmp_vec r2svv(const tmp_vec& v)
	{
		return r2sv(v.x, v.y, v.z);
	}

	 ~tmp_vec() {}
};

typedef tmp_vec<realnum> Vector;

extern wchar_t vctr_xcom_opr[];
extern wchar_t vctr_ycom_opr[];
extern wchar_t vctr_zcom_opr[];
	
extern bool unitvctr;


inline std::wostream& operator<<(std::wostream& os, const Vector& v)
{
		if(unitvctr)
		{
			if(v.x != 0.0) os<<v.x<<vctr_xcom_opr;
			
			if(v.y != 0.0) 
			{
				if(v.x != 0.0) os<<std::showpos;
				os<<v.y<<vctr_ycom_opr;
			}

			if(v.z != 0.0)
			{
				if(v.y != 0.0) os<<std::showpos;
				os<<v.z<<vctr_zcom_opr;
			}
			
			os<<std::noshowpos;

			if(v.x == 0 && v.y == 0 && v.z == 0)
				os<<_T("(0, 0, 0)");

		}
		else
		{
			os<<std::noshowpos;
			os<<_T("(")<<v.x<<_T(", ")<<v.y<<_T(", ")<<v.z<<_T(")");
		}
	
		return os;
}