#pragma once

#include <cmath>
#include <iostream>
#include <ios>

typedef struct { bool scalar; double x, y, z; } Vector;

extern wchar_t vctr_xcom_opr[];
extern wchar_t vctr_ycom_opr[];
extern wchar_t vctr_zcom_opr[];
	
extern bool unitvctr;

inline std::wostream& operator<<(std::wostream& os, const Vector& v)
{
		if(unitvctr)
		{
			if(v.x != 0.0) os<<v.x<<vctr_xcom_opr;
			
			if(v.y != 0.0) 
			{
				if(v.x != 0.0) os<<std::showpos;
				os<<v.y<<vctr_ycom_opr;
			}

			if(v.z != 0.0)
			{
				if(v.y != 0.0) os<<std::showpos;
				os<<v.z<<vctr_zcom_opr;
			}
			
			os<<std::noshowpos;

			if(v.x == 0 && v.y == 0 && v.z == 0)
				os<<_T("(0, 0, 0)");

		}
		else
		{
			os<<std::noshowpos;
			os<<_T("(")<<v.x<<_T(", ")<<v.y<<_T(", ")<<v.z<<_T(")");
		}
	
		return os;
}

// constructor
inline Vector MakeVector(double i=0, double j=0, double k=0, bool is_scalar=false)
{
	Vector v; v.scalar = is_scalar;
	v.x = i; v.y = j; v.z = k;
	return v;
}

inline Vector operator+(const Vector& v1, const Vector& v2)
{
	return MakeVector(v1.x+v2.x, v1.y+v2.y, v1.z+v2.z);
}

inline Vector operator-(const Vector& v1, const Vector& v2)
{
	return MakeVector(v1.x-v2.x, v1.y-v2.y, v1.z-v2.z);
}

// unary minus
inline Vector operator-(const Vector& v)
{
	return MakeVector(-v.x, -v.y, -v.z, v.scalar);
}

// cross product
inline Vector operator%(const Vector& v1, const Vector& v2)
{
	return MakeVector(v1.y*v2.z - v1.z*v2.y,
					  v1.z*v2.x - v1.x*v2.z,
					  v1.x*v2.y - v1.y*v2.x);
}

// dot product
inline double operator*(const Vector& v1, const Vector& v2)
{
	return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z;
}

// scalar * vector
inline Vector operator*(double d, const Vector& v)
{
	return MakeVector(d*v.x, d*v.y, d*v.z);
}

// vector * scalar
inline Vector operator*(const Vector& v, double d)
{
	return MakeVector(v.x*d, v.y*d, v.z*d);
}

// vector / scalar
inline Vector operator/(const Vector& v, double d)
{
	return MakeVector(v.x/d, v.y/d, v.z/d);
}

// absolute value
inline double absv(const Vector& v)
{
	return sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
}

// norm value
inline double normv(const Vector& v)
{
	return v.x*v.x+v.y*v.y+v.z*v.z;
}

inline bool operator > (const Vector& v1, const Vector& v2)
{
	return normv(v1)>normv(v2);
}

inline bool operator < (const Vector& v1, const Vector& v2)
{
	return normv(v1)<normv(v2);
}

inline bool operator == (const Vector& v1, const Vector& v2)
{
	return (v1.x==v2.x)&&(v1.y==v2.y)&&(v1.z==v2.z);
}

inline bool operator != (const Vector& v1, const Vector& v2)
{
	return !(v1==v2);
}

inline bool operator >= (const Vector& v1, const Vector& v2)
{
	double n1 = normv(v1), n2 = normv(v2);
	if(n1>n2)
		return true;
	else if(v1 == v2)
		return true;
	else
		return false;
}

inline bool operator <= (const Vector& v1, const Vector& v2)
{
	double n1 = normv(v1), n2 = normv(v2);
	if(n1<n2)
		return true;
	else if(v1 == v2)
		return true;
	else
		return false;
}

inline double xcomv(const Vector& v)
{
	return v.x;
}

inline double ycomv(const Vector& v)
{
	return v.y;
}

inline double zcomv(const Vector& v)
{
	return v.z;
}


// cylindrical to rectangular
inline Vector c2rv(double rho, double psi, double z)
{
	return MakeVector(rho*cos(psi), rho*sin(psi), z); 
}

// cylindrical to rectangler
inline Vector r2cv(double x, double y, double z)
{
	double p = sqrt(x*x+y*y);
	double psi;
	double mpi2 = M_PI/2.0;
	
	if(p == 0.0)
		psi = 0.0;
	else
	{
		if(x==0)
		{
			if(y>0.0)
				psi = mpi2;
			else // y<0.0
				psi = -mpi2;
		}
		else
			psi = atan(y/x);
	}

	return MakeVector(p, psi, z);
}

// sperical to rectanglur
inline Vector s2rv(double r, double theta, double psi)
{
	double p = r*sin(theta);
	return MakeVector(p*cos(psi), p*sin(psi), r*cos(theta));
}

// rectangular to spherical
inline Vector r2sv(double x, double y, double z)
{
	double p = x*x+y*y;
	double r = sqrt(p+z*z);
	double theta;
	double psi;

	double mpi2 = M_PI/2.0;

	if(z == 0.0)
		theta = mpi2;
	else
		theta = atan(sqrt(p)/z);

	if(x == 0.0)
	{
		if(y>0.0)
			psi = mpi2;
		else if(y<0.0)
			psi = -mpi2;
		else
			psi = 0;
	}
	else
		psi = atan(y/x);

	return MakeVector(r, theta, psi);
}


// cylindrical to rectangular
inline Vector c2rvv(const Vector& v)
{
	return c2rv(v.x, v.y, v.z);
}

// cylindrical to rectangler
inline Vector r2cvv(const Vector& v)
{
	return r2cv(v.x, v.y, v.z);
}

// sperical to rectanglur
inline Vector s2rvv(const Vector& v)
{
	return s2rv(v.x, v.y, v.z);
}

// rectangular to spherical
inline Vector r2svv(const Vector& v)
{
	return r2sv(v.x, v.y, v.z);
}
